/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of suplib
 *
 * suplib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * suplib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdlib.h>
#include <string.h>

#include "str/str.h"
#include "debug/debug.h"

static int charptr_comp(const void *s1, const void *s2);

#define DELIM ",: "

static size_t nflags;
static char **flag_list = NULL;
static char *flags = NULL;

/*

  PROCEDURE
  debug_init - parse debug flags and initialize internals

  DESCRIPTION
  @code{debug_init} and @code{dbflag} comprise a simple system
  for allowing users to specify debug flags to a program in
  a human understandable fashion.

  @code{debug_init} parses a string for debug flags.  The flags
  are comma separated tokens of any characters (preferrably
  alphanumeric).  It builds an internal list of the specified
  flags, which can be tested with the @code{dbflag} function.

  There is but a single list kept per process, and subsequent
  calls to @code{debug_init} will overwrite the list.

  To free the memory associated with the internal list,
  @code{debug_init} should be called with @var{debug_flags}
  set to @code{NULL}.

  RETURNS
  It returns zero upon success, nonzero if it ran out of memory

  AUTHOR
  Diab Jerius
*/
int
debug_init(
  const char *debug_flags	/* a comma delimited list of debug flags */
  )
{
  char *start, **f;

  if ( flags ) free( flags );
  if ( flag_list ) free( flag_list );

  flags = NULL;
  flag_list = NULL;

  /* determine number of flags */
 if ( debug_flags && (nflags = str_tokcnt( debug_flags, DELIM )) )
  {
    if ( NULL == ( flags = str_dup( debug_flags ) ) )
      return 1;

    if ( NULL == (flag_list = (char **) malloc(nflags * sizeof(char *))) )
    {
      free( flags );
      flags = NULL;
      return 1;
    }

    start = strtok(flags, DELIM);
    f = flag_list;
    while (start)
    {
      *f++ = start;
      start = strtok(NULL, DELIM);
    };

    qsort( flag_list, nflags, sizeof(char *), charptr_comp);
  }

  return 0;
}

/*

  PROCEDURE
  dbflag - test for a debug flag

  DESCRIPTION
  @code{dbflag} tests if the passed string was in the list of debug flags
  previously parsed by @code{debug_init}.

  RETURNS
  It returns zero if the flag was not specified, non-zero if it was.

  AUTHOR
  Diab Jerius

*/
int
dbflag(
  const char *dbf		/* the debug flag to test */
  )
{
  return nflags &&
    bsearch(&dbf, flag_list, nflags, sizeof(char *), charptr_comp);
}

static int
charptr_comp(const void *s1, const void *s2)
{
  return strcmp(*((const char **)s1), *((const char **)s2));
}
