/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of suplib
 *
 * suplib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * suplib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "file/file.h"

#define DIRSEP '/'
#define DIRSEP_S "/"
#define PATHSEP ':'

static char *
copy_string( char *string )
{
  char *copy = (char *) malloc( strlen( string ) + 1);
  if ( copy )
    strcpy( copy, string );

  return copy;
}

/*

  PROCEDURE
  searchpath - Search a set of paths for a file.

  DESCRIPTION
  @code{searchpath} scans a list of paths for a file.  The paths are
  specified in a string, separated by the @code{PATHSEP} character
  (`:' for UNIX). The paths are searched in the order that they occur in
  the string.  If an empty path is found in the specification, a user
  specified default path (or the current directory, should this not be
  specified) will be searched at that point.  If the filename
  begins with @code{DIRSEP} (`/' for UNIX), it is used directly.

  RETURNS
  It returns a pointer to a dynamically allocated string holding the
  complete path, or @code{NULL} if it couldn't find the file or it
  ran out of memory.  The calling procedure must free this string.

  AUTHOR
  Diab Jerius
*/
char *
searchpath(
  char *file,			/* the file to look for */
  char *default_path,		/* the default path */
  char *path_spec		/* the list of paths */
)
{
  char *path = NULL;
  char *path_s, *path_s0;

  if ( NULL == file || ! *file )
    return NULL;

  if ( file[0] == DIRSEP )
    return access( file, F_OK ) ? NULL : copy_string( file );

  if ( NULL == path_spec )
    path_spec = ":";

  /* if default path doesn't exist, use current directory */
  if ( NULL == default_path )
    default_path = "";

  /* need to diddle with path string, copy it in case it's read only */
  if ( NULL == ( path_s = copy_string( path_spec ) ) )
    return NULL;

  path_s0 = path_s;

  for ( ;; )
  {
    char *start = path_s;

    while ( *path_s && *path_s != PATHSEP )
      path_s++;

    if ( start == path_s )	/* empty path spec, use default */
    {
      path = (char *) malloc( strlen( default_path ) + strlen( file ) + 2 );

      /* if default path is empty, don't stick a DIRSEP in there. */
      if ( strlen( default_path ) > 0 )
        strcat( strcat( strcpy( path, default_path ), DIRSEP_S ), file );
      else
        strcpy( path, file );
    }
    else
    {
      size_t len_path = path_s - start;
      path = (char *) malloc( len_path + strlen( file ) + 2 );
      strncpy( path, start, path_s - start );
      path[len_path] = '\0';
      strcat( strcat( path, DIRSEP_S ), file );
    }

    if ( ! access( path, F_OK ) )
      break;

    free( path );
    path = NULL;

    if ( *path_s )
       path_s++;
    else
      break;
  }

  free( path_s0 );

  return path;
}
