/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of suplib
 *
 * suplib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * suplib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stddef.h>
#include <math.h>

#include "imagefcts/imagefcts.h"


/*
  PROCEDURE

  ave_dev_err - determine statistics for a list of objects.

  DESCRIPTION
  @code{ave_dev_err} determines the total weight, weighted average,
  uncertainty in weighted average, weighted deviation and uncertainty
  in weighted deviation for a list of objects.

  AUTHOR
  Diab Jerius
*/

void
ave_dev_err(
  void *objs, /*  pointer to list of objects to process */
  size_t n_obj, /*  number of objects in list */
  size_t s_obj, /* size of an object, in bytes */
  double *x_ave,  /* weighted average of objects */
  double *x_ave_err, /*  uncertainty in weighted average */
  double *x_dev, /*  weighted deviation of objects */
  double *x_dev_err, /*  uncertainty in weighted deviation */
  double *tot_wt, /*  total weight of objects */
  void (*get_stuff)(const void *obj, double *x, double *x_err,
                                        double *w, double *w_err)  /* pointer
		    to function which returns the weight, position,
		    and uncertainties in weight and position of an
		    object */
  )
{
  double f_n;
  double dev, var = 0.0, wt = 0.0, ave = 0.0, ave_err2 = 0.0,
              dev_err, dev_err1 = 0.0, dev_err2 = 0.0, dev_err3 = 0.0;
  char *obj;
  char *objlim = (char *) objs + n_obj * s_obj;

  /* calculate average */
  for (obj=objs; obj < objlim ; obj += s_obj)
  {
    double x, x_err, w, w_err;

    get_stuff(obj, &x, &x_err, &w, &w_err);
    ave += w * x;
    wt += w;
  }
  ave /= wt;


  /* calculate errors */
  ave_err2 = 0.0;

  for (obj=objs; obj < objlim ; obj += s_obj)
  {
    double x, x_err, x_err2, w, w2, w_err, w_err2, dx, dx2;

    get_stuff(obj, &x, &x_err, &w, &w_err);

    x_err2 = x_err * x_err;
    w_err2 = w_err * w_err;
    w2 = w * w;

    dx = x - ave;
    dx2 = dx * dx;

    ave_err2 += (w2 * x_err2 + dx2 * w_err2) / wt / wt;

    dev_err1 += dx2 * (4 * w2 * x_err2 + dx2 * w_err2) / wt / wt;
    dev_err2 += dx2 * w_err2 / wt / wt;
    dev_err3 += w_err2 / wt / wt;

    var += w * dx2 / wt;
  }

  f_n = (double) n_obj / (n_obj - 1);
  var *= f_n;
  dev = sqrt(var);
  dev_err = sqrt(f_n * f_n / var * dev_err1 - 2 * f_n * dev_err2 + var * dev_err3) / 2;

  *x_ave = ave;
  *x_ave_err = sqrt(ave_err2);
  *x_dev = dev;
  *x_dev_err = dev_err;
  *tot_wt = wt;
}


/*
    @example
  @{
    \def\sumwi@{\sum_i w_i@}
    \def\dd#1#2@{@{\partial #1 \over \partial #2@}@}
    \def\dmudxi@{\dd@{\mu@}@{x_i@}@}
    \def\dmudwi@{\dd@{\mu@}@{w_i@}@}

    \def\nnm@{\left( @{n \over n-1@} \right )@}
    \def\dsdxi@{\dd@{s@}@{x_i@}@}
    \def\dsdwi@{\dd@{s@}@{w_i@}@}
    \def\dsdqi@{\dd@{s@}@{q_i@}@}
    \def\dsdmu@{\dd@{s@}@{\mu@}@}
    \def\dsigsqdqi@{\dd@{\sigma^2@}@{q_i@}@}
    \def\dsigsqdxi@{\dd@{\sigma^2@}@{x_i@}@}
    \def\dsigsqdwi@{\dd@{\sigma^2@}@{w_i@}@}
    \def\dsigsqdmu@{\dd@{\sigma^2@}@{\mu@}@}

    $$\eqalign@{%
    \mu &= @{\sum_i w_i x_i \over \sum_i w_i@}\; ; \quad
    \sigma_\mu^2 = \sum_i \left ( \dmudxi \right )^2 \sigma_@{x_i@}^2 +
		      \left ( \dmudwi \right )^2 \sigma_@{w_i@}^2\cr
    \dmudxi &= @{w_i \over \sumwi@}\; ; \quad
    \dmudwi = @{x_i \over \sumwi@} -
		    @{\sumwi x_i \over \left ( \sumwi \right)^2@} =
	    @{x_i \over \sumwi@} - @{\mu \over \sumwi@} =
	       @{x_i - \mu \over \sumwi@} \cr
    \sigma_\mu^2 &=  @{1 \over \left (\sumwi \right )^2@}
		     \sum_i \left [ w_i^2 \sigma_@{x_i@}^2
				+ (x_i - \mu)^2 \sigma_@{w_i@}^2 \right ]\cr
    \cr
    s &= \sqrt@{\sigma^2@} \; ; \quad
	  \sigma^2 = @{\sum_i w_i (x_i - \mu)^2 \over \sumwi@} \nnm\cr
    \sigma_s^2 &= \left ( \dsdmu \right )^2 \sigma_@{\mu@}^2  +
		  \sum_i \left [ \left ( \dsdxi \right )^2 \sigma_@{x_i@}^2 +
		      \left ( \dsdwi \right )^2 \sigma_@{w_i@}^2 \right ] \; ;
			   \quad \dsdqi = @{1 \over 2\sigma@}\dsigsqdqi\cr
    \dsigsqdmu &= -@{\sum_i 2w_i(x_i - \mu) \over \sumwi@} \nnm\cr
	       &= -2\left ( @{\sum_i w_i x_i \over \sumwi@} - 
			    \mu @{\sumwi \over \sumwi@} \right) \nnm
		= -2\left ( \mu - \mu \right ) \nnm = 0\cr
    \dsigsqdxi &= @{2w_i(x_i - \mu) \over \sumwi@} \nnm\cr
    \dsigsqdwi &= \left ( @{(x_i - \mu)^2 \over \sumwi@} -
			  @{\sum_i w_i (x_i - \mu)^2 \over
			      \left (\sumwi \right)^2@} \right )
			  \nnm =
		  @{1 \over \sumwi@}
		       \left ( (x_i - \mu)^2 \nnm  - \sigma^2 \right ) \cr
    \sigma_s^2 &= @{1 \over 4\sigma^2 (\sumwi)^2@} 
		   \sum_i \left [ 4 w_i^2(x_i - \mu)^2 \nnm^2 \sigma_@{x_i@}^2 +
		      \left ( (x_i - \mu)^2 \nnm  - \sigma^2 \right )^2
			     \sigma_@{w_i@}^2 \right ]\cr
	       &= @{1 \over 4\sigma^2 (\sumwi)^2@} \sum_i
		    \left [
		      4 w_i^2(x_i - \mu)^2 \nnm^2 \sigma_@{x_i@}^2 +
			     (x_i - \mu)^4 \nnm^2 \sigma_@{w_i@}^2
		    \right ]\cr
	       &\quad +
		  @{1 \over 4\sigma^2 (\sumwi)^2@} \sum_i
		      -2 (x_i - \mu)^2 \nnm \sigma^2 \sigma_@{w_i@}^2 \cr
	       &\quad +
		  @{1 \over 4\sigma^2 (\sumwi)^2@} \sum_i
		      \sigma^4 \sigma_@{w_i@}^2 \cr
	       &= @{1 \over 4(\sumwi)^2@} \Biggl [
		      @{1 \over \sigma^2@} \nnm^2 \sum_i (x_i - \mu)^2 \left [
		      4w_i^2 \sigma_@{x_i@}^2 + (x_i - \mu)^2 \sigma_@{w_i@}^2
		      \right ] \cr
	       &\qquad\qquad\qquad -
		  2 \nnm \sum_i
		      (x_i - \mu)^2 \sigma_@{w_i@}^2
		+
		  \sigma^2 \sum_i \sigma_@{w_i@}^2
		\Biggr ] \cr
    @}$$
  @}

  @end example
*/
