/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of suplib
 *
 * suplib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * suplib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdlib.h>
#include <stddef.h>
#include <math.h>

#include "lists/lists.h"
#include "imagefcts/imagefcts.h"

static void reduce_var(char *obj, char *obj_lim, size_t s_obj,
                       unsigned long *n, double x0, double *var,
                       double *tot_wt, double (*get_x)(const void *obj, double *x));

/*
  PROCEDURE
  center_variter -  iteratively determine the center of a distribution

  DESCRIPTION
  @code{center_variter} determines the center of a distribution by
  iteratively rejecting objects whose deviation in distance from a
  determined center is greater than a given number of sample
  deviations from the center.  After each iteration, changes in the
  determined center are measured; if the changes are less than a
  specified threshold, the algorithm is deemed to have converged.
  Additionally, there is a limit on the number of iterations
  performed.  The input list of objects is reordered.

  A more detailed description of the algorithm:

  @enumerate

  @item
  a mean center is determined and used as the initial guess

  @item
  distances to the determined center are calculated for all objects,
  and the variance (with respect to zero distance) is calculated.

  @item
  objects with variances greater than the specified limit are rejected
  and the group variance recalculated after each rejection.  this is
  repeated until either no more objects are rejected, the object count
  drops below a limit, or an iteration limit is reached.

  @item
  a new mean center is calculated using the objects surviving the
  culling, and compared to the previous determination.  if the
  distance between the two centers is less than the specified
  tolerance, the procedure ends.  if the iteration limit has not been
  reached, the process continues with step 2.

  @end enumerate
  
  RETURNS
  It returns the following information about the set of objects which
  survived the clipping spree:

  @table @code

  @item center
  the final determined center

  @item dev_used
  the mean deviation from the center

  @item n_used
  the number of objects

  @item wt_used
  the weight of the objects

  @item objs_used
  a pointer into the (reordered) input list where the
  final list of objects starts

  @end table

  The actual value returned by @code{center_variter} will be one of:

  @table @code

  @item CVR_OK
  everything went well

  @item CVR_LESSTHANTWO
  less than two objects were left after clipping.  @code{fvar} was
  probably too low

  @item CVR_MAXITER
  the iteration limit was reached.

  @end table

  AUTHOR
  Diab Jerius

*/

int
center_variter(
  void *objs,			/* the list of objects to process */
  void *wwork,			/* a work space of size 3 * s_obj */
  unsigned long n_objs,		/* total number of objects to process */
  double tot_wt,		/* the total (summed) weight of all of the
				   objects. */
  size_t s_obj,			/* the size of an object in bytes */
  double dtol,			/* smallest absolute (not percentage) difference
				   between two distances so as to consider them
				   distinct. */
  double fvar,			/* fraction of determined variance in distance above
				   which to ignore objects in center determination. */
  unsigned long max_iter,	/*  maximum number of iterations to perform. */
  unsigned long max_clip,	/*  maximum number of clips per iteration to perform. */
  double *center,  		/* the final determined center */
  double *dev_used,		/* the standard deviation of the
				   objects remaining after the last round of clips */
  size_t *n_used,		/* the number of objects remaining
				   after the last round of clips */
  double *wt_used,		/* the summed weights of the objects
				   remaining after the last round of
				   clips */
  void **objs_used,		/* a pointer to the objects used */
  double (*get_x)(const void *obj, double *x), /*  routine which retrieves position of an object. returns weight of
						   object */
  void (*put_x)(void *obj, double x), 	/*  routine which stuffs a position into an object. */
  int (*comp)(const void *obj1, const void *obj2) /*  routine which compares two objects based upon
							  their squared distances (same setup as routines
							  for *<qsort>*) */
  )
{
  double x;
  unsigned long iter;
  double xcent;
  double var = 0.0;
  double wt = tot_wt;
  unsigned long n = n_objs;
  char *p = NULL;

  /* determine initial center */
  xcent = weightpos(objs, n_objs, tot_wt, s_obj, get_x);

  /* outermost iteration over centers */
  for (iter=0 ; max_iter ? iter < max_iter : 1 ; iter++)
  {
    double diff;

    p = objs;
    n = n_objs;
    wt = tot_wt;
    /* determine initial unnormalized group variance */
    var = wtvar(xcent, objs, n_objs, tot_wt, s_obj, get_x);

    /* if everything's at one point, why bother? */
    if (var > 0.0 && n > 2)
    {
      unsigned long n_old;
      unsigned long clip = 0;
      void *p_obj = wwork;
      void *work = (char *) wwork + s_obj;

      do
      {
        void *t_obj;

        n_old = n;

        /* partition object list about minimum acceptable x */
        if (tot_wt)
          put_x(p_obj, xcent - sqrt(fvar * (var / wt) * ((double) n / (n-1)) ));
        else
          put_x(p_obj, xcent - sqrt(fvar * (var / (n-1)) ));

        t_obj = partition(p, p_obj, work, s_obj, n, comp);

	if (t_obj)
	{
	  reduce_var(p, t_obj, s_obj, &n, xcent, &var, &wt, get_x);
	  p = t_obj;
	}

        /* partition object list about maximum acceptable x */
        if (tot_wt)
          put_x(p_obj,
                    xcent + sqrt(fvar * (var / wt) * ((double) n / (n-1)) ));
        else
          put_x(p_obj, xcent + sqrt(fvar * (var / (n-1)) ));

        t_obj = partition(p, p_obj, work, s_obj, n, comp);
	if (t_obj)
          reduce_var((char *) t_obj + s_obj, p + n * s_obj, s_obj, &n, xcent, &var, &wt, get_x);

      } while (n != n_old  && n > 1 && ++clip != max_clip);

      if (n < 2)
        return CVR_LESSTHANTWO;
    }

    x = weightpos(p, n, wt, s_obj, get_x);

    diff = (x - xcent);
    if (diff < dtol) break;

    xcent = x;
  }

  if (max_iter && iter == max_iter)
    return CVR_MAXITER;

  *center = xcent;
  if (n > 1)
    *dev_used =  tot_wt ? sqrt((var / wt) * ((double) n / (n - 1))) :
			  sqrt(var / (n - 1));
  else
    *dev_used = 0;
  *n_used = n;
  *wt_used = wt;

  *objs_used = p;

  return CVR_OK;
}


static void
reduce_var(
  char *obj,
  char *obj_lim,
  size_t s_obj,
  unsigned long *n,
  double x0,
  double *var,
  double *tot_wt,
  double (*get_x)(const void *obj, double *x)
  )
{
  if (tot_wt)
    for (; obj < obj_lim ; obj += s_obj)
    {
      double x;
      double wt = get_x(obj, &x);
      double dx = x0 - x;

      *var -= wt * dx * dx;
      *tot_wt -= wt;
      (*n)--;
    }
  else
    for (; obj < obj_lim ; obj += s_obj)
    {
      double x;
      double dx;

      get_x(obj, &x);
      dx = x0 - x;

      *var -= dx * dx;
      (*n)--;
    }
}
