/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of suplib
 *
 * suplib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * suplib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include "io/io.h"

typedef struct
{
  char *buf;
  size_t len;
  size_t extend;
} FGetRecord;

/*

  PROCEDURE
  fget_rec_new - Create a @code{fget_rec} buffer structure.

  DESCRIPTION
  This routine creates a buffer structure that can be used by
  @code{fget_rec}.  The buffer structure should be destroyed with
  @code{fget_rec_delete}.

  The calling procedure must provide the initial size of the buffer
  (including the trailing '\0' which seals off the string),
  as well as the amount by which the buffer should be extended if the
  buffer is shorter than the input record.  @code{fget_rec} repeatedly
  extends the buffer until a record fits, so this value needn't be
  that large.  Note that the @code{buf_size} argument should be
  greater than @samp{1}.

  RETURNS
  It returns a pointer to a buffer structure (@emph{not} the actual
  buffer) or @code{NULL} if it was unable to allocate it.

  AUTHOR
  Diab Jerius

 */
void *
fget_rec_new(
  size_t buf_size,	/* the initial size of the buffer */
  size_t extend		/* the amount to extend the buffer each time
			   it is too small */
  )
{
  FGetRecord *rec = (FGetRecord *) malloc( sizeof(FGetRecord) );

  /* force the initial buffer size to have room for at least one
     input character as well as the EOS character */
  if ( buf_size < 2 )
    buf_size = 2;

  if ( rec )
  {
    if ( NULL == ( rec->buf = (char *) malloc( buf_size ) ) )
      free( rec );
    else
    {
      rec->len = buf_size;
      rec->extend = extend;
      rec->buf[0] = '\0';	/* ensure that it's a valid string */
    }
  }

  return rec;
}

/*

  PROCEDURE
  fget_rec_delete - Delete a @code{fget_rec} buffer structure.

  DESCRIPTION
  This routine deletes a buffer structure previously created by
  @code{fget_rec_new}.  It also deletes the accompanying buffer.

  AUTHOR
  Diab Jerius

 */
void
fget_rec_delete(
  void *rec		/* an @code{fget_rec} buffer structure to delete */
  )
{
  if ( rec )
  {
    free( ((FGetRecord *) rec)->buf );
    free( rec );
  }
}

/*

  PROCEDURE
  fget_rec_read - Read a record of arbitrary length.

  DESCRIPTION
  @code{fget_rec_read} reads data from the passed stream until either an
  end of line character or the end of file is reached.  It removes any
  end of line character from the string.  It writes the data into a
  user provided buffer, and extends the buffer as necessary to hold a
  complete input record.  The end of the buffer is terminated with a
  '\0'.  The buffer structure should created by @code{fget_rec_new}
  and deleted with @code{fget_rec_delete}.  The data is written into 
  the buffer at the position specified by the @code{start} parameter.
  If the starting position is beyond the end of the buffer the buffer 
  is extended and the string currently in the buffer is extended with
  spaces to have length @code{start}. This happens even if there is no 
  new data available in the input stream.

  RETURNS
  It normally returns a pointer to the data that has been read.
  Upon end of file, it returns @code{NULL}.  If it cannot extend
  the record, it returns @code{NULL} and sets @code{errno} to
  @code{ENOMEM}.  If there was a read error on the stream, it
  returns @code{NULL} and the stream's error flag is set (use
  @code{ferror} to check it).

  AUTHOR
  Diab Jerius

 */
char *
fget_rec_read(
  FILE *fin,		/* the input stream from which to read the data */
  void *recbuf,		/* an @code{fget_rec} buffer structure, created by
			   @code{fget_rec_new} */
  size_t start		/* the zero-based index into the buffer where the
			   incoming record should be stored */
)
{
  FGetRecord *rec = (FGetRecord *) recbuf;
  char *s;
  size_t len = 0;
  int nbytes;

  /* if the starting position is beyond the end of the buffer (taking
     into account the EOS character), extend it and fill the gap in
     with spaces */
  if ( start >= rec->len - 1)
  {
    size_t len = strlen( rec->buf );
    char *t = (char *) realloc( rec->buf, start + 1 );

    if ( NULL == t )
    {
      errno = ENOMEM;
      return NULL;
    }

    rec->len = start + 1;
    rec->buf = t;
    memset( rec->buf + len, ' ', start - len);
    rec->buf[start] = '\0';
  }

  nbytes = rec->len - start;
  for ( len = start, s = rec->buf + start ;
	fgets( s, nbytes, fin ) ;
	s = rec->buf + len)
  {
    len += strlen( s );
    if ( '\n' == rec->buf[len - 1] )
    {
      /* end of line encountered, return string */
      rec->buf[len - 1] = '\0';
      return rec->buf;
    }
    else
    {
      /*
	if length of string is equal to length requested, then there's
	probably more coming, so extend the buffer.  if not, it's
	probably at EOF with no '\n', so loop again, just to be
	official.
       */
      if ( len == rec->len - 1 )
      {
	/* must extend array */
	char *t = (char *) realloc( rec->buf, rec->len + rec->extend );

	if ( NULL == t )
	{
	  errno = ENOMEM;
	  return NULL;
	}
	rec->len += rec->extend;
	rec->buf = t;
	s = rec->buf + len;
	nbytes = rec->extend + 1;
      }
    }
  }
  
  return ( s != rec->buf + start ) ? rec->buf : NULL;
}


/*

  PROCEDURE
  fget_rec - Read a record of arbitrary length.

  DESCRIPTION
  @code{fget_rec} reads data from the passed stream until either an
  end of line character or the end of file is reached.  It removes any
  end of line character from the string.  It writes the data into a
  user provided buffer, and extends the buffer as necessary to hold a
  complete input record.  The end of the buffer is terminated with a
  '\0'.  The buffer structure should created by @code{fget_rec_new}
  and deleted with @code{fget_rec_delete}.

  This is essentially a wrapper around @code{fget_rec_read}, with
  @code{start = 0}.


  RETURNS
  It normally returns a pointer to the data that has been read.
  Upon end of file, it returns @code{NULL}.  If it cannot extend
  the record, it returns @code{NULL} and sets @code{errno} to
  @code{ENOMEM}.  If there was a read error on the stream, it
  returns @code{NULL} and the stream's error flag is set (use
  @code{ferror} to check it).

  AUTHOR
  Diab Jerius

 */
char *
fget_rec(
  FILE *fin,		/* the input stream from which to read the data */
  void *recbuf		/* an @code{fget_rec} buffer structure, created by
			   @code{fget_rec_new} */
)
{
  return fget_rec_read( fin, recbuf, 0 );
}

/*

  PROCEDURE
  fget_rec_append - Append a record of arbitrary length.

  DESCRIPTION
  @code{fget_rec_append} reads data from the passed stream until either an
  end of line character or the end of file is reached.  It removes any
  end of line character from the string.  It appends the data to the
  end of the data in the user provided buffer, and extends the buffer
  as necessary to hold a complete input record.  The end of the buffer
  is terminated with a '\0'.  The buffer structure should created by
  @code{fget_rec_new} and deleted with @code{fget_rec_delete}.  Note
  that it appends the new data to the end of the

  This is essentially a wrapper around @code{fget_rec_read}, with
  @code{start = strlen(buf)}.


  RETURNS
  It normally returns a pointer to the data that has been read.
  Upon end of file, it returns @code{NULL}.  If it cannot extend
  the record, it returns @code{NULL} and sets @code{errno} to
  @code{ENOMEM}.  If there was a read error on the stream, it
  returns @code{NULL} and the stream's error flag is set (use
  @code{ferror} to check it).

  AUTHOR
  Diab Jerius

 */
char *
fget_rec_append(
  FILE *fin,		/* the input stream from which to read the data */
  void *recbuf		/* an @code{fget_rec} buffer structure, created by
			   @code{fget_rec_new} */
)
{
  FGetRecord *rec = (FGetRecord *) recbuf;
  size_t start = strlen( rec->buf );
  return fget_rec_read( fin, recbuf, start );
}



#ifdef TEST

int
main( int argc, char *argv[] )
{
  char *buf;
  void *rec = fget_rec_new( 2, 1 );
  FILE *file = fopen( __FILE__, "r" );

  if ( NULL == rec )
  {
    fputs( "couldn't fget_rec_new\n", stderr);
    exit( 1 );
  }

  errno = 0;

#if 0
  while( NULL != ( buf = fget_rec( file, rec ) ) )
    puts( buf );

  if ( ENOMEM == errno )
  {
    fputs( "couldn't extend record\n", stderr );
    exit( 1 );
  }
    
  if ( ferror( file ) )
  {
    fputs( "error on input stream\n", stderr );
    exit( 1 );
  }

  rewind(file);

#endif
  
  buf = fget_rec( file, rec );
  while( buf )
  {
    puts( buf );
    buf = fget_rec_append( file, rec );
  }

  if ( ENOMEM == errno )
  {
    fputs( "couldn't extend record\n", stderr );
    exit( 1 );
  }
    
  if ( ferror( file ) )
  {
    fputs( "error on input stream\n", stderr );
    exit( 1 );
  }
  fget_rec_delete( rec );
}
#endif
