/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of suplib
 *
 * suplib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * suplib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdlib.h>
#include <errno.h>
#include <string.h>

#include "str/str.h"
#include "keyval/keyval.h"

static TokListToken boolstr[] =
{
  { "0",	0 },
  { "1",	1 },
  { "no",	0 },
  { "off",	0 },
  { "on",	1 },
  { "yes",	1 },
};

static TokList boolstrs = GenTokList( boolstr );

static int kv_cmp( const void *v1, const void *v2 );

#define ERROR(val) do { error = KeyValErr_##val; goto error; } while(0)
#define WHERE(ptr) (ptr - spec)

#define DELIM	";, \t"

/*

  PROCEDURE
  keyval_st - parse a string with keyword-value pairs

  DESCRIPTION

  @code{keyval_st} parses a string which contains multiple keyword-value
  pairs, or boolean switches.  Pairs or switches should be separated
  by semi-colons, commas, spaces, or tabs.  Keywords and values are
  separated by an equals sign.  Values which are strings with embedded
  spaces should be surrounded by qutoes.  To embed a quote, escape it
  with @samp{\}. Boolean values may be specified either as
  keyword-value pairs, or just as the keyword (indicating @var{true})
  with an optional @samp{!} ( indicating @var{false}).  In the former
  case, the value may be any of @samp{yes}, @samp{no}, @samp{0},
  @samp{1}, @samp{on}, @samp{off} (case is not significant).

  The resultant values are stored in a user supplied structure.  Offsets
  into the structure are encoded in an array of @code{KeyVal} structures,
  which my be constructed by the caller. The structures have the
  following definition:

  @example
  typedef struct KeyVal
  @{
    char *key;		keyword name
    KeyType type;	keyword type
    size_t offset;	offset of value
    int set;		set to true if keyword read
    int (*xfrm)( char *in, void *out );
  @} KeyVal;
  @end example

  The @var{type} field is one of @code{Key_String},
  @code{Key_Integer}, @code{Key_Float}, @code{Key_Double}, or
  @code{Key_Boolean}. The @var{offset} is the byte offset into the
  user supplied structure where the value is to be stored.  The
  @var{set} field is set by @code{keyval_st} if that keyword was
  present in the passed specification string.

  The @var{xfrm} field points to an optional function which takes the
  value string (@var{in}) and converts it to the correct type.  It
  should store the converted value in the address pointed to by the
  @var{out} argument.  For string types, it should store a freshly
  allocated string.  It should return zero upon success, non-zero upon
  failure.  @var{xfrm} should be set to @code{NULL} if not used.  It
  is not used for @code{Key_Boolean} types.

  The keyword types and their storage requirements are as follows:

  @table @code

  @item Key_String
  @code{keyval} will allocate space for a string; in this case, the
  field in the structure must be a @code{char *} so as to accept a
  pointer to the string.  If the contents of that pointer are not
  @code{NULL}, @code{keyval} will free that memory first.

  @item Key_Integer
  the field must be an @code{int};

  @item Key_Float
  the field must be a @code{float};

  @item Key_Double
  the field must be a @code{double};

  @item Key_Boolean
  the field must be an @code{int};

  @end table
  
  Construction of this table is eased by use of the
  @code{KeyValStEntry} preprocessor macro:

  @example

  where to stick the values
  typedef struct Data
  @{
    int	n_flies;
    double fly_lifetime;
    char *food;
  @} Data;

  KeyVal vals[] =
  @{
    KeyValStEntry( fly_lifetime, Key_Double, Data ),
    KeyValStEntry( food, Key_String, Data ),
    KeyValStEntry( n_flies, Key_Integer, Data ),
  @};
  #define NFIELDS ( sizeof(vals) / sizeof(KeyVal ) )

  @end example

  The @code{KeyValStEntry} macro assumes that the name of the keyword
  is the same as the name of the field in the structure which is to
  receive the data.  You can get around this restriction by specifying
  the values to the @code{KeyVal} structure directly.

  The entries in the @code{KeyVal} array @emph{must} be in alphabetical order
  by keyword name. (See @code{keyval_cmp}.)

  RETURNS
  It returns a code indicating whether an error ocurred. (See
  @file{suplib/keyval.h} for the possible errors, and codes.  It also
  returns the location of the error in the passed specification via
  the passed @code{where} variable.  The @code{keyval_perror} routine
  will print a human understandable form of a @code{KeyValErr}.

  AUTHOR
  Diab Jerius

 */

KeyValErr
keyval_st(
  const char *	kv_spec,/* keyword - value specification string to parse */
  KeyVal *	map,	/* keyword value type and offset map */
  size_t	nkey,	/* number of keywords in map */
  void *	data,	/* where the data are to be stored */
  long *	where	/* position in string where error occurred */
)
{
  char *ptr, dchar, *tp;
  void *dp;
  char *token = NULL;
  KeyVal *kv;
  char *spec;
  KeyValErr error;

  *where = 0;

  errno = 0;
  
  if ( NULL == ( spec = str_dup( kv_spec ) ) )
    ERROR( NOMEM );

  /* split up specification into quoted chunks */
  token = str_tokq( spec, DELIM, 1, 1, &ptr, &dchar );
  if ( NULL == token && EINVAL == errno )
    ERROR( UNBAL );


  while( token )
  {
    /* keyword value pair */
    if ( (tp = strchr( token, '=' )) )
    {
      *tp = '\0';
      if ( NULL ==
	   ( kv = bsearch( token, map, nkey, sizeof(KeyVal), kv_cmp ) ) )
	ERROR( NOKEY );

      /* remove possible quotes from value */
      token = tp + 1;
      if ( NULL == unquote( token ) )
	ERROR( UNBAL );
      /* turn escaped characters into the real mccoy */
      if ( NULL == unescape( token ) )
	ERROR( UNBAL );
      
      dp = (char *) data + kv->offset;
      if ( kv->xfrm )
      {
	if ( Key_String == kv->type && * (char **) dp ) free ( * (char **) dp );
	if ( kv->xfrm( token, dp ) )
	  ERROR( INVAL );
      }
      else
      {
	switch ( kv->type )
	{
	case Key_String:
	  if ( * (char **) dp )
	    free( * (char **) dp );
	  if ( NULL == ( tp = str_dup( token ) ) )
	    ERROR( NOMEM );
	  * (char **) dp = tp;
	  break;

	case Key_Integer:
	  {
	    int val;
	    errno = 0;
	    val = strtol( token, &tp, 10 );
	    if ( *tp )
	      ERROR( INVAL );
	    if ( errno )
	      ERROR( RANGE );

	    * (int *) dp = val;
	  }
	  break;

	case Key_Float:
	  {
	    double val;
	    errno = 0;
	    val = strtod( token, &tp );
	    if ( *tp )
	      ERROR( INVAL );
	    if ( errno )
	      ERROR( RANGE );

	    * (float *) dp = val;
	  }
	  break;

	case Key_Double:
	  {
	    double val;
	    errno = 0;
	    val = strtod( token, &tp );
	    if ( *tp )
	      ERROR( INVAL );
	    if ( errno )
	      ERROR( RANGE );

	    * (double *) dp = val;
	  }
	  break;

	case Key_Boolean:
	  {
	    int val;
	    /* humor them and test against the standard boolean type flags */

	    if ( -1 == ( val = tokcasematch( token, &boolstrs ) ) )
	      ERROR( INVAL );

	    * (int *) dp = val;
	  }
	  break;
	}
      }
    }

    /* boolean */
    else
    {
      int val;
      /* check for leading `!', which indicates negation */
      val = ( '!' == token[0] ) ? 0 : 1;

      if ( NULL ==
	   ( kv = bsearch( token + (1 - val), map, nkey, sizeof(KeyVal), kv_cmp ) ) )
	ERROR( NOKEY );

      if ( Key_Boolean != kv->type )
	ERROR( INVAL );

      * (int *) dp = val;
    }
    
    kv->set = 1;
    token = str_tokq( NULL, DELIM, 1, 1, &ptr, &dchar );

    if ( NULL == token && EINVAL == errno )
      ERROR( UNBAL );

  }

  if ( spec )
    free( spec );
  return  KeyValErr_OK;

 error:
  if ( spec )
    free( spec );
  if ( token )
    *where = WHERE( token );
  return error;
}


static int
kv_cmp( const void *v1, const void *v2 )
{
  return strcmp( v1, ((KeyVal *) v2)->key );
}


/*
  
  PROCEDURE
  keyval_cmp - compare @code{KeyVal} structures, alphabetically

  DESCRIPTION
  @code{keyval_cmp} is an alphabetical comparison routine for
  @code{KeyVal} structures intended for use by @code{bsearch} or
  @code{qsort}.
  
  RETURNS
  It returns @samp{-1} if the first structure is lexically earlier than the
  second, @samp{0} if they are the same, and @samp{1} if the first is
  lexically after the second.

  AUTHOR
  Diab Jerius

*/

int
keyval_cmp(
  const void *v1,	/* pointer to first @code{KeyVal} structure */
  const void *v2	/* pointer to second @code{KeyVal} structure */
)
{
  return strcmp( ((KeyVal *) v1)->key, ((KeyVal *) v2)->key );
}
