/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of suplib
 *
 * suplib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * suplib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stddef.h>
#include <stdio.h>


#include "lists/lists.h"

/*

  PROCEDURE
    bnd_bsearch -  binary search with bounding element return

  DESCRIPTION
   @code{bnd_bsearch} performs a binary search upon a list of objects,
   returning the matching object or the bounding objects if the key
   is not found.

   @itemize @bullet

   @item 
   If an object matches @var{key}, its address is returned and
   @var{lo_bnd} and @var{hi_bnd} are set to the address.

   @item
   If an object is not found, @code{NULL} is returned, and
   @var{hi_bnd} and @var{lo_bnd} are set as follows:

   @itemize @bullet

   @item

   If @var{key} falls between two objects, @var{lo_bnd} and
   @var{hi_bnd} are set to the addresses of the bracketing objects.

   @item

   If @var{key} is less than any object in the list, @var{lo_bnd} is
   set to @code{NULL} and @var{hi_bnd} is set to @var{base}.

   @item

   If @var{key} is larger than any object in the list, @var{lo_bnd} is
   set to @var{base} + (@var{n}-1) and @var{hi_bnd} is set to
   @code{NULL}.

   @end itemize
   @end itemize

  WARNING
   Note that while @var{hi_bnd} and @var{lo_bnd} are declared as @code{void *},
   the calling routine @emph{must} pass a pointer to a pointer!

  AUTHOR
    Diab Jerius

*/
const void *
bnd_bsearch(
  const void *key, 	/* a pointer to the comparison key */
  const void *base,	/* the list of objects */
  size_t n,		/* the number of objects in the list */
  size_t size,		/* the size of an object in bytes */
  const void **lo_bnd,	/* the address of a pointer which will be set to
                           the address of the largest object lower than
                           the key */
  const void **hi_bnd,	/* the address of a pointer which will be set to
                           the address of the smallest object higher than
                           the key */
  int (*cmp) (const void *keyval, const void *datum)
   			/* a routine which compares the user supplied key to an
                           object, returning <1, 0, or >1 if the key is
		           respectively less than, equal to, or greater
			   than the object */
)
{
  void *lo_ptr, *hi_ptr, *ptr;
  int comp; 

  size_t hi = n - 1;
  size_t lo = 0;
  size_t mid = (hi + lo) >> 1;

  if (n == 0)
  {
    *lo_bnd = NULL;
    *hi_bnd = NULL;
    return NULL;
  }

  if ( n == 1 )
  {
    comp = cmp(key, base);

    if ( comp < 0 )
    {
      *lo_bnd = NULL;
      *hi_bnd = base;
      return NULL;
    }

    if ( comp > 0 )
    {
      *lo_bnd = base;
      *hi_bnd = NULL;
      return NULL;
    }

    *lo_bnd = *hi_bnd = base;
    return base;
  }

  /* if only two elements, mid = lo, which breaks when key < lo */
  if ( n == 2 && cmp( key, base ) < 0 )
      mid = hi;


  while ( hi - lo > 1 )
  {

    /* find midpoint in list. integer divide by 2, rounding down */
    mid = (hi + lo) >> 1;

    ptr = (char *) base + mid * size;
    comp = cmp(key, ptr);
    if (comp < 0)
      hi = mid;
    else if (comp > 0)
      lo = mid;
    else
    {
      *lo_bnd = ptr;
      *hi_bnd = ptr;
      return ptr;
    }
  }

  hi_ptr = (char *) base + hi * size;
  lo_ptr = (char *) base + lo * size;

  /* if lo == mid, previous comparison indicated that mid was too low;
     check hi */
  if ( lo == mid )
  {
    comp = cmp(key, hi_ptr );

    if ( 0 == comp )
    {
      *lo_bnd = hi_ptr;
      *hi_bnd = hi_ptr;
      return hi_ptr;
    }

    if ( comp > 0 )
    {
      *lo_bnd = (char *) base + (n-1) * size;
      *hi_bnd = NULL;
    }

    else
    {
      *lo_bnd = lo_ptr;
      *hi_bnd = hi_ptr;
    }
  }
  else
  {
    comp = cmp( key, lo_ptr );

    if ( 0 == comp )
    {
      *lo_bnd = lo_ptr;
      *hi_bnd = lo_ptr;
      return lo_ptr;
    }

    if ( comp < 0 )
    {
      *lo_bnd = NULL;
      *hi_bnd = (void *) base;
    }
    else
    {
      *lo_bnd = lo_ptr;
      *hi_bnd = hi_ptr;
    }
  }


  return NULL;
}
