/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of suplib
 *
 * suplib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * suplib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include <string.h>
#include <errno.h>

#include "str/str.h"

#include "range/range.h"

#define DELIM " ,\t"

#define ERROR(val) do { error = val; goto error; } while(0)
#define WHERE(ptr) (ptr - range_str)

static int	rangef_bucket_cmp( const void *e1, const void *e2 );
static int	rangef_start_comp(const void *r1, const void *r2);

/*

  PROCEDURE
  rangef_parse - parse a list of floating point ranges

  DESCRIPTION
  @code{rangef_parse} translate a list of ranges of the form
  "r1,r2,r3,r4" into start-end pairs. a range has the format:

  @example
     [start]@{ `:' | `(' | `)' | `~' @}[end]
  @end example

  where start and end are optional limits on the range.

  the range operators are as follows:

  @table @samp

  @item :
  range from start to end, inclusive of both.

  @item (
  range from start to end, exclusive of end.

  @item )
  range from start to end, exclusive of start.

  @item ~
  range from start to end, exclusive of both.

  @end table

The handling of the
  ranges depends upon the flags set in the @var{opts} arguement.
  @var{opts} is set to the logical OR of zero or more of the following values

  @table @code

  @vindex Range_SORT
  @item Range_SORT
  The resultant ranges will be sorted in ascending order by their lower
  bound

  @vindex Range_MERGE
  @item Range_MERGE
  Adjacent ranges will be merged, if possible.
  This implies @code{Range_SORT}.

  @vindex Range_INCOMPLETE
  @item Range_INCOMPLETE
  If set, incomplete ranges, i.e., those without an explicit start or end
  value, are accepted.  In this case, implied starts are set to the
  value of the passed @var{minval} argument, while implied ends are set to
  the value of the @var{maxval} argument.

  @end table

  @code{rangef_parse} creates a @code{RangeFList} object and returns it
  via the @var{rfl} parameter.

  RETURNS
  It returns a code indicating whether an error ocurred. (See
  @file{suplib/range.h} for the possible errors, and codes.  It also
  returns the location of the error in the range specification via
  the passed @code{where} variable.

  AUTHOR
  Diab Jerius

*/

RangeErr
rangef_parse(
  RangeFList **rfl,	/* returned list of ranges			*/
  char *range_spec,	/* string containing range to parse		*/
  RangeOpts opts,	/* range options				*/
  double minval,	/* replacement value for incomplete start	*/
  double maxval,	/* replacement value for incomplete end		*/
  long *where		/* position in string where error occurred	*/
)
{
  char *ptr, *tok, *range_str;
  RangeFList *rl;
  RangeF *range;
  RangeErr error = RangeErr_OK;
  size_t nr;

  *rfl = NULL;
  *where = 0;


  /* determine number of ranges */
  if ((nr = str_tokcnt(range_spec, DELIM)) == 0)
  {
    if ( NULL == (rl = rangef_new( 0 )) )
      return RangeErr_NOMEM;
    return RangeErr_OK;
  }

  if ( NULL == (rl = rangef_new( nr )) )
    return RangeErr_NOMEM;

  if ( NULL == ( range_str = str_dup(range_spec) ) )
  {
    free(rl);
    return RangeErr_NOMEM;
  }

  if ( ! opts & Range_INCOMPLETE )
  {
    minval = RANGEF_MIN;
    maxval = RANGEF_MAX;
  }

  range = rl->range;
  for (tok = strtok(range_str, DELIM);
       tok ;
       tok = strtok(NULL, DELIM), range++)
  {
    size_t p, op, start, end;

    /* split token into start; op; end; */
    p = strcspn(tok, ":()~");
    op = tok[p];

    /*
       determine if start or end limits have been specified.  if so,
       set flags to that effect; if not, preset limits to minval and
       maxval
    */

    if (p > 0)
    {
      start = 1;
    }
    else
    {
      start = 0;
      range->start = minval;
    }

    if (p < strlen(tok) - 1)
    {
      end = 1;
    }
    else
    {
      end = 0;
      range->end = maxval;
    }

    /* split token into start and end limits */
    tok[p] = '\0';

    /* if incomplete range specified and not authorized, flag as error */
    if (op && (! (opts & Range_INCOMPLETE) && (!start || !end)) )
      ERROR( RangeErr_INCOMPLETE );

    if (start)
    {
      errno = 0;
      range->start = strtod(tok, &ptr);

      if (errno)
	ERROR( RangeErr_ERANGE );

      if (*ptr != '\0')
	ERROR( RangeErr_ILLNUM );
    }

    /* reconstruct range for error output */
    tok[p] = op;

    if (end)
    {
      errno = 0;
      range->end = strtod(tok=tok+p+1, &ptr);

      if (errno)
	ERROR( RangeErr_ERANGE );

      if (*ptr != '\0')
	ERROR( RangeErr_ILLNUM );
    }

    /* out of order limits are an error */
    if (range->end < range->start)
	ERROR( RangeErr_ORDER );

    switch (op)
    {
      case ':':
        range->op = RangeOp_INCLUSIVE;
        break;

      case ')':
        range->op = RangeOp_START;
        break;

      case '(':
        range->op = RangeOp_END;
        break;

      case '~':
        range->op = RangeOp_START | RangeOp_END;
        break;
    }

    if ( range->op != RangeOp_INCLUSIVE &&
	range->start == range->end)
      ERROR( RangeErr_EMPTY );
  }


  /* sort ranges by starting value */
  if ( opts & Range_SORT || opts & Range_MERGE )
  {
    qsort(rl->range, nr, sizeof(RangeF), rangef_start_comp);

    if ( opts & Range_MERGE )
    {
      RangeF
	*r0 = rl->range,
	*r1 = rl->range + 1;

      /*
	 in most cases, will be able to perform binary search after
	 a merge.  however, if can't guarantee unique range buckets,
	 can't do binary search
      */
      rl->bsearch = 1;
      for (; r1 < rl->range + nr ; r1++)
      {
        if (r1->start > r0->end)
	{
	  r0++;
	  if (r0 != r1)
	    *r0 = *r1;
	}	
        else
        {
          /* extra special problems if both start's are the same */
          if (r0->start == r1->start)
          {
            /* set start op to least restrictive */
            if (! (r1->op & RangeOp_START) )
              r0->op &= ~RangeOp_START;
          }
	  /* if the ranges overlap, but both are exclusive, skip */
          else if (r1->start == r0->end &&
		   (r0->op & RangeOp_END && r1->op & RangeOp_START))
	  {
	    r0++;
	    if (r0 != r1)
	      *r0 = *r1;
	    /* can't do binary search on this */
	    rl->bsearch = 0;
	    continue;
          }

          /* if the second range is longer, copy to first range */
          if (r1->end > r0->end)
          {
            r0->op |= r1->op & RangeOp_END;
            r0->end = r1->end;
          }
          /* if both ranges have same length, worry about restrictions */
          else if (r1->end == r0->end)
          {
            if (! (r1->op & RangeOp_END) )
              r0->op &= ~RangeOp_END;
          }
        }
      }
      nr = (r0 - rl->range) + 1;

      if ((r0 = (RangeF *) realloc(rl->range, nr * sizeof(RangeF))) == NULL)
	ERROR( RangeErr_NOMEM );
      rl->range = r0;
    }
  }

  rl->n = nr;

  free( range_str );
  *rfl = rl;
  return RangeErr_OK;

error:
  rangef_del( rl );
  free( range_str );
  *where = WHERE(tok);
  return error;
}


/*

  PROCEDURE
  rangef_new - create a new @code{RangeFList}

  DESCRIPTION
  @code{RangeFList} objects are used to contain a list of integral ranges.
  @code{rangel_new} creates such an object for a given number of ranges.
  
  RETURNS
  It returns a pointer to a @code{RangeFList} object upon success,
  @code{NULL} if it ran out of memory.

  AUTHOR
  Diab Jerius

*/
RangeFList *
rangef_new( size_t nr )
{
  RangeFList *rl = NULL;

  if ( (rl = (RangeFList *) malloc(sizeof(RangeFList))) )
  {
    if ( NULL ==
	 (rl->range = (RangeF *) malloc(nr * sizeof(RangeF))) )
    {
      free(rl);
      rl = NULL;
    }
    else
    {
      rl->n = 0;
      rl->bsearch = 0;
    }
  }

  return rl;
}


/*

  PROCEDURE
  rangef_del - destroy a @code{RangeFList} object.

  DESCRIPTION
  This routine frees all memory associated with a @code{RangeFList} object.
  
  RETURNS
  It returns @code{NULL}.

  AUTHOR
  Diab Jerius

*/
RangeFList *
rangef_del( RangeFList *rl )
{
  if ( rl )
  {
    if ( rl->range )
      free( rl->range );
    free( rl );
  }
  return NULL;
}


/*

  PROCEDURE
  rangef_in - determine if a number is within a list of ranges
  
  DESCRIPTION
  @code{rangef_in} determines if a passed number is encompassed
  by one of the ranges in a passed @code{RangeFList}.

  RETURNS
  It returns non-zero if the number is within the list of ranges,
  zero if not.

  AUTHOR
  Diab Jerius

*/  
int
rangef_in(
  RangeFList *rl,
  double t
  )
{
  if ( rl->bsearch )
  {
    return NULL != bsearch( &t, rl->range, rl->n, sizeof(RangeF),
			    rangef_bucket_cmp );
  }
  else
  {
    RangeF *range = rl->range;
    RangeF *rnglim = range + rl->n;
    int start, end;

    for (start = end = 0; !(start && end) && range < rnglim; range++)
    {
      start = (range->op & RangeOp_START) ?
               (t > range->start)  :
               (t >= range->start) ;

      end = (range->op & RangeOp_END) ?
	       (t < range->end)  :
               (t <= range->end) ;
    }
    return start && end;
  }
}

static int
rangef_bucket_cmp( const void *e1, const void *e2 )
{
  double t = *((double *) e1);
  RangeF *range = (RangeF *) e2;

  return 
    (
      (range->op & RangeOp_START) ?
      (t > range->start)  :
      (t >= range->start)
    )
    &&
    (
      (range->op & RangeOp_END) ?
      (t < range->end)  :
      (t <= range->end)
    );
}

static int
rangef_start_comp(const void *r1, const void *r2)
{
  return ((RangeF *)r1)->start - ((RangeF *)r2)->start;
}

