/*
 *
 *  The original routines upon which these are based are available
 *  here:
 *
 *    http://lib.stat.cmu.edu/apstat/
 *    http://www.ams.sunysb.edu/~nkfr/5PUBL.HTM
 *
 *  and were published in Applied Statistics, Vol. 43, No. 1 (1994), 265-270.
 *  "Algorithm AS 288: Exact Smirnov Two-Sample Tests
 *     for Arbitrary Distributions"
 *
 *  The http://lib.stat.cmu.edu/apstat/ code is copyrighted by the
 *  Royal Statistical Society. It may be distributed provided that no
 *  fee is charged.
 */

#include <locale.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>

#include "stats.h"

#include "local.h"


/*

  PROCEDURE
  fl - Calculation of the natural logarithm of I!

  DESCRIPTION
  For 1 <= i <= 19 the precise values stored in the array GLN are 
  used (calculated recursively using extended, 80-bit reals of 
  Borland Turbo Pascal 5.5 compiler).

  For larger values of I the formula 6.1.41 from M.Abramowitz
  and I.Stegun (1964) for the logarithm of gamma-function is used
  (with the expansion to the seventh degree of I+1).

  AUTHOR
  Original by Andrei M. Nikiforov
  C transcription by Diab Jerius

  COPYRIGHT
  This code is copyrighted by the Royal Statistical Society. It may be
  distributed provided that no fee is charged.
*/

static double gln[] =
  { 0.0,
    0.693147180559945309,      1.791759469228055000,
    3.178053830347945620,      4.787491742782045990,
    6.579251212010101000,      8.525161361065414300,
    10.604602902745250200,     12.801827480081469600,
    15.104412573075515300,     17.502307845873885800,
    19.987214495661886200,     22.552163853123422900,
    25.191221182738681500,     27.899271383840891600,
    30.671860106080672800,     33.505073450136888900,
    36.395445208033053600,     39.339884187199494000
  } ;

#define R12   ( 1.0 /   12)
#define R360  (-1.0 /  360)
#define R1260 ( 1.0 / 1260)
#define R1680 (-1.0 / 1680)
#define SLN2PI 0.918938533204673

static double fl(int i)
{
  if (i < 19)
    return gln[i-1];
  else
  {
    double z = i + 1;
    double z2 = z * z;
    return
      (z - 0.5) * log(z)
      + ( SLN2PI + (R12 + (R360 + (R1260 + R1680 / z2) / z2) /  z2) / z) - z;
  }
}

/*

  PROCEDURE
  gsmirn - exact Smirnov Two-Sample tests for arbitrary distributions.

  DESCRIPTION
  @code{gsmirn} generates the P-value for the generalized two-sample
  Smirnov tests.  It calculates the probability of the null hypothesis
  (that the two samples are the same) based upon one of three
  statistics

  @enumerate

  @item
  sup | X - Y |
  
  @item
  sup ( X - Y )

  @item
  sup ( Y - X )

  @end enumerate

  The input parameter @code{kind} indicates which of these should be
  calculated.  The input parameters @code{m} and @code{dstat} are
  calculated by the @code{stcalc} subroutine.

  See Applied Statistics, Vol. 43, No. 1 (1994), 265-270.

  WARNING
  This code returns incorrect values for large @code{nx} and @code{ny}
  for @code{nx != ny} for various ratios.  See @code{gsmirn2} for a
  slower version which does not suffer from these problems.

  RETURNS
  @table @code

  @item 0
  no error

  @item 1
  the @code{nx} < 1 or @code{ny} < 1

  @item 2
  @code{kind} != 1, 2 or 3

  @item 3
  @code{q} is not positive

  @item 4
  @code{m} is inconsistent with @code{nx} and @code{ny} or has
  non-positive elements

  @item 5
  allocation of workspace failed

  @end table


  AUTHOR
  Original by Andrei M. Nikiforov
  C transcription by Diab Jerius

*/

#define EPS 	1E-6
#define SMALL	1E-35
#define SMALLN  -80.5904782547916
#define ALN2	0.69314718056
#define CHKNUM  1e32
#define ITERUP  116

int gsmirn(
  int nx,	/* The number of observations in the first sample */
  int ny,	/* The number of observations in the second sample */
  int kind,	/* The hypothesis tested */
  int *m,	/* The number of observations falling into each of K
		   categories (with ascending order of category values) */
  double dstat, /* The statistic */
  double *q	/* output: p-value */
  )
{
  int ii, itmp;
  double dtmp;

  /* Local variables */
  int l, n, l2, ic;
  double dl;

  int ile;
  int iri, ile2, iri2, icat, iles, iris, jlow, jupp;
  int ileft;
  int nties;

  double scl, slope, delta, deviat;

  int nofdiv, iright;

  /* work array */
  double *p, *P;

  int newrct;

  /* allocate work array */
  P = (double *) malloc( (nx + 2) *  sizeof(double));
  if ( NULL == P )
    return 5;

  m--;
  p = P - 1;

  /* Function Body */
  n = nx + ny;

  if (nx <= 0 || ny <= 0) 
    return 1;

  if (kind < 1 || kind > 3)
  {
    free(P);
    return 2;
  }

  for( l=1, icat=0; icat < n ; l++ )
  {
    if (m[l] <= 0)
    {
      free(P);
      return 4;
    }

    icat += m[l];

    if ( icat > n )
    {
      free(P);
      return 4;
    }
  }

  *q = 1.0;
  delta = dstat - EPS;
  if (delta <= 0.0)
  {
    free(P);
    return 0;
  }

  p[1] = 1.0;

  /*  Parameters to define a set for trajectories to lie within it */
  slope = nx / (double) n;
  deviat = slope * delta * ny;
  /* delta = deviat; */

  newrct = 1;
  icat = 1;
  nties = m[1];
  ile = 0;
  iri = 0;

  /*  Variables to prevent from overflows in P */
  ic = ITERUP;
  nofdiv = 0;
  scl = 1.0;

  for (l = 1; l <= n - 1; l++)
  {
    if (nties == 1)
    {

      /*  Calculate boundaries for current L (if Lth value in the */
      /*  pooled sample is unique or `last' in a series of tied values) */
      dl = l * slope;

      /* t = (double) l / n */
      /* deviat = delta * sqrt (t * (1.0 - t)) */

      IMIN3(iri, (int) (dl + deviat), l, nx );
      IMAX3(ile, (int) (dl - deviat + 1.0), l - ny, 0 );
      icat++;
      nties = m[icat];
      newrct = 1;

    }
    else
    {
      /*      Calculations for tied observations */
      nties--;

      /*  If we have the first observation with the new value (that */
      /*  is not unique), then determine the ICATth `rectangle' */
      if (newrct) {
	newrct = 0;

	l2 = l + nties;
	/*
	  if (l2 <= n)
	  {
	    iri2 = nx;
	    ile2 = nx;
	  }  else
	*/
	dl = l2 * slope;

	/* t = (double) l2 / n;
	   deviat = delta * sqrt (t * (1.0 - t));
	*/


	/*  X axis boundaries of the subset of `line' L(l+1) within
	    ICATth rectangle */

	IMIN3(iri2, (int) (dl + deviat), l2, nx );
	IMAX3(ile2, (int) (dl - deviat + 1.0), l2 - ny, 0 );

	/* } */

	/*  Four sides of the rectangle on the X and Y axes */
	ileft = ile;
	iright = iri2;
	jupp = l2 - ile2;
	jlow = l - iri - 1;
      }

      /*  Calculate boundaries for current L (Lth value is tied) */
      IMAX2( ile, ileft,  l - jupp );
      IMIN2( iri, iright, l - jlow );
    }

    /*  Set the left (right) boundary for the one-sided test */

    switch( kind )
    {
    case 2:
      itmp = l - ny;
      ile = MAX(0, itmp);
      break;

    case 3:
      iri = MIN(nx,l);
      break;
    }

    /*  Calculate the number of trajectories p(i,j) for current L */
    iles = MAX(1,ile);

    itmp = l - 1;
    iris = MIN( itmp, iri );

    for ( ii = iris; ii >= iles; ii-- )
      p[ii + 1] += p[ii];


    /*  Check whether elements of P are large enough to multiply them
	by SMALL */

    ic--;
    if (ic <= 0)
    {
      dl = 0.0;

      for (ii = iles + 1; ii <= iris + 1 ; ii++)
	if ( dl < p[ii] ) dl = p[ii];

      if (dl == 0.0)
      {
	free(P);
	return 0;
      }

      if (dl > CHKNUM)
      {
	for (ii = iles + 1 ; ii <= iris + 1; ii++)
	  p[ii] *= SMALL;

	ic = ITERUP;
	nofdiv++;
	scl *= SMALL;
      }
      else
      {
	/*  Estimate the number of iterations for DL=Pmax to became of
	    order 1/SMALL */

	ic = (-SMALLN - log(dl)) / ALN2;
      }
    }

    /*  Define, whether boundaries lie on the left and lower sides of
	the rectangle R and define boundary values for the next
	iterations */

    if (ile == 0) 
      p[iles] = scl;

    else 
      p[iles] = 0.0;

    if (iri == l) 
      p[iris + 2] = scl;
    else
      p[iris + 2] = 0.0;

  }

  dl = p[nx + 1] + p[nx];

  /* get rid of work array */
  free(P);

  if (dl == 0.0)
    return 0;

  /* The P-value */
  *q = 1.0 - exp(fl(nx) + fl(ny) + log(dl) - nofdiv * SMALLN - fl(n));

  /*  Q=1 is allowable, Q<=0 is not (accuracy loss due to rounding errors) */
  if (*q <= 0.0)
    return 3;

  return 0;
}



