/*
 *
 *  The original routine upon which this is based is available
 *  here:
 *
 *  http://www.ams.sunysb.edu/~nkfr/ARCH/AL288.ZIP
 *
 *  As there is no copyright or license notice, it is assumed to be freely
 *  redistributable with no restrictions.  
 *
 *  See also gsmirn.c
 * */


#include <locale.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>

#include "stats.h"

#include "local.h"


/*

  PROCEDURE
  gsmirn2 - exact Smirnov Two-Sample tests for arbitrary distributions.

  DESCRIPTION
  @code{gsmirn2} generates the P-value for the generalized two-sample
  Smirnov tests.  It calculates the probability of the null hypothesis
  (that the two samples are the same) based upon one of three
  statistics

  @enumerate

  @item
  sup | X - Y |
  
  @item
  sup ( X - Y )

  @item
  sup ( Y - X )

  @end enumerate

  The input parameter @code{kind} indicates which of these should be
  calculated.  The input parameters @code{m} and @code{dstat} are
  calculated by the @code{stcalc} subroutine.

  This version uses the method of Timonin and Chernomordik (Theor.
  Prob.Appl.,1985) to avoid large numbers for P instead of direct
  scaling adopted by published algorithm AS 288.

  One-sided tests are coded separately to use the fast calculation scheme

  See also @code{gsmirn}.

  RETURNS
  @table @code

  @item 0
  no error

  @item 1
  the @code{nx} < 1 or @code{ny} < 1

  @item 2
  @code{kind} != 1, 2 or 3

  @item 3
  @code{q} is not positive

  @item 4
  @code{m} is inconsistent with @code{nx} and @code{ny} or has
  non-positive elements

  @item 5
  allocation of workspace failed

  @end table


  AUTHOR
  Original by Andrei M. Nikiforov
  C transcription by Diab Jerius

  COPYRIGHT
  As there is no copyright or license notice, it is assumed to be freely
  redistributable with no restrictions.  
*/

#define EPS 	1E-6

int gsmirn2(
  int nx,	/* The number of observations in the first sample */
  int ny,	/* The number of observations in the second sample */
  int kind,	/* The hypothesis tested */
  int *m,	/* The number of observations falling into each of K
		   categories (with ascending order of category values) */
  double dstat, /* The statistic */
  double *q	/* output: p-value */
  )
{

  int i, n, l, icat, newrct, nties, iri, iris, ile, iles, jupp, jlow;
  double slope, delta, deviat, devm1, dl, ple, pri;

  /* work array */
  double *p, *P;

  /* Swap Nx and Ny to invert the set for the alternative
     'F(x) > G(x)' */

  if ( 2 == kind )
  {
    int l = nx;
    nx = ny;
    ny = l;
  }

  /* allocate work array */
  P = (double *) malloc( (nx + 2) *  sizeof(double));
  if ( NULL == P )
    return 5;

  m--;
  p = P - 1;

    /* Function Body */
  n = nx + ny;

  if (nx <= 0 || ny <= 0) 
  {
    free(P);
    return 1;
  }

  if (kind < 1 || kind > 3)
  {
    free(P);
    return 2;
  }

  for( l=1, icat=0; icat < n ; l++ )
  {
    if (m[l] <= 0)
    {
      free(P);
      return 4;
    }

    icat += m[l];

    if ( icat > n )
    {
      free(P);
      return 4;
    }
  }

  delta = dstat - EPS;
  if (delta <= 0.0)
  {
    *q = 1.0;
    free(P);
    return 0;
  }


  slope = nx / (double) n;
  deviat = slope * delta * ny;
  devm1 = deviat - 1.0;
  ile = 0;
  newrct = 1;
  icat = 1;
  nties = m[1];

  /*  Two - sided Test */
  if ( kind == 1 )
  {
    int iright = 0;
    int ileft = 0;

    iri = 0;
    p[1] = 1.0;

    for ( l = 1; l <= n-1 ; l++ )
    {
      if (nties == 1)
      {

	/* Calculate boundaries for current L (if L - th value in the
	   pooled sample is unique or `last' in a series of tied
	   values) */

	dl = l * slope;
	IMIN3(iri, (dl + deviat), l, nx);
	IMAX3(ile, (dl - devm1), l - ny, 0);
	icat++;
	nties = m[icat];
	newrct = 1;
      }
      else
      {
	/* Calculations for tied observations */

	nties = nties - 1;

	/* If we have the first observation with the new value
	   (that is not unique), then determine the ICAT-th `rectangle' */
	
	if (newrct)
	{
	  int l2, ile2;

	  newrct = 0;
	  l2 = l + nties;
	  dl = l2 * slope;

	  /* X axis boundaries of the subset of `line' L(l+1) within
	     ICAT-th rectangle */

	  IMAX3(ile2, (dl - devm1), l2 - ny, 0);
	  IMIN3(iright, (dl + deviat), l2, nx);

	  /* Right side of the rectangle is IRIGHT. Find three other
	     sides. */
	  ileft = ile;
	  jupp = l2 - ile2;
	  jlow = l - iri - 1;
	}

	/*  Calculate boundaries for current L (L - th value is tied) */

	IMAX2(ile, ileft, l - jupp);
	IMIN2(iri, iright, l - jlow);
      }

      /* Define, whether boundaries lie on the left and lower sides of the
	 rectangle R */

      if (ile == 0)
      {
	ple = 1.0;
	iles = 1;
      }
      else
      {
	ple = 0.0;
	iles = ile;
      }

      if (iri == l)
      {
	pri = 1.0;
	iris = iri - 1;
      }
      else
      {
	pri = 0.0;
	iris = iri;
      }

      /* Calculate probabilities p(i,j) for current L */
      dl = 1.0 / l;

      for ( i = iris; i >= iles ; i-- )
	p[i + 1] = (p[i + 1] * (l - i) + p[i] * i) * dl;

      p[iles] = ple;
      p[iris + 2] = pri;

    }


  }

  /* One - sided Test */
  else
  {
    int ileft = 0;
    int jmax = 0;

    for ( i = 1 ; i <= nx + 1 ; i++ )
      p[i] = 1.0;

    for ( l = 1; l <= n-1 ; l++ )
    {
      if (nties == 1)
      {
	IMAX3(ile, (l * slope - devm1), l - ny, 0);
	icat++;
	nties = m[icat];
	newrct = 1;
      }
      else
      {
	nties--;
	if (newrct)
	{
	  int l2, ltmp;

	  newrct = 0;
	  l2 = l + nties;
	  ileft = ile;
	  IMAX3(ltmp, (l2 * slope - devm1), l2 - ny, 0);
	  jupp = l2 - ltmp;
	}
	IMAX2(ile, ileft, l - jupp);
      }

      if (ile == 0)
      {
	/*  find the highest point jmax */
	jmax = l;
      }
      else
      {
	/* right (lower) boundary is as large as possible */
	IMIN2(iri, nx, l - jmax);
	dl = 1.0 / l;
	for ( i = iri; i >= ile; i-- )
	  p[i + 1] = (p[i + 1] * (l - i) + p[i] * i) * dl;
	p[ile] = 0.0;
      }
    }
  }

  *q = 1.0 - (p[nx + 1] * ny + p[nx] * nx) / n;

  free(P);

  /* q=1 is allowable, q<=0 is not (accuracy loss due to rounding
     errors) */

  if (*q <= 0.0)
    return 3;

  return 0;
}
