/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of suplib
 *
 * suplib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * suplib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>

#include "str/str.h"


/*   
  PROCEDURE
  str_tokenize - tokenize a string at specified delimiters

  DESCRIPTION
  This routine tokenizes the string using @code{str_tokq}.  The 
  invoking routine calls @code{str_tokenize} once to parse all the tokens
  for a given delimiter.  The tokens are stored in @emph{string_argv}.  

  if the @code{restore} argument is false, @code{str_tokenize} will
  allocate memory for each of the tokens; otherwise it will return
  pointers to the locations in the original string.

  The routine @code{str_tokenize_free} is provided to free the memory
  allocated by @code{str_tokenize}

  RETURNS
  The number of tokens. The third argument, @code{string_argv}, 
  contains the parsed tokens.

  ERRORS
  On error, @code{str_tokenize} returns @code{0} and sets @code{errno}
  accordingly.   The following errors are recognized:

  @table @code

  @item EINVAL
    There were unbalanced quotes or the last character in the
    string was an escape character (i.e., no character to escape).

  @item ENOMEM
    A memory allocation failed

  @end table

  AUTHOR
  Dan Nguyen

  Diab Jerius
*/
int str_tokenize(
  char* string,			/* string to parse */
  const char* delim,		/* the delimiters to use */
  char*** string_argv,		/* to store the parsed tokens */
  int skip,    /* true if consecutive delimiters are treated as
		   a single delimiter */
  int restore  /* true if @code{str_tokq} should restore the
				  delimiter character from the previous 
				  invocation */
  ) {

  char *tok, *ptr, dchar;
  char **string_arr;
  int n_tokens;

  *string_argv = NULL;
  n_tokens = str_tokqcnt( string, delim, skip );
  
  /* if there was an error, or no tokens, return */
  if ( n_tokens <= 0 ) 
    return 0;

  string_arr = *string_argv = (char**) malloc( (n_tokens + 1) * sizeof(char*));

  /* errno is set to ENOMEM by malloc */
  if ( NULL == string_arr )
    return 0;
  
  /* no need to check for errors in str_tokq, as str_tokqcnt has
     parsed the string once already */
  
  for ( tok = str_tokq( string, delim, skip, restore, &ptr, &dchar ) ;
	tok ;
	tok = str_tokq( NULL,   delim, skip, restore, &ptr, &dchar ),
	string_arr++
	)
  {
    if ( 0 == restore )
      *string_arr = tok;

    else
    {
      *string_arr = str_dup( tok );

      /* out of memory. must delete everything now */
      if ( NULL == *string_arr )
      {
	str_tokenize_free( *string_argv, restore );
	errno = ENOMEM;
	return 0;
      }
    }

  }

  /* last element is a sentinel */
  *string_arr = (char*) NULL;
  
  return n_tokens;

}

/*   
  PROCEDURE
  str_tokenize_free - free memory allocated by @code{str_tokenize}

  DESCRIPTION
  This routine frees the memory allocated by the @code{str_tokenize}.
  The user must call @code{str_tokenize_free} with the same value for
  restore as was used when @code{str_tokenize} was invoked.
  
  RETURNS
  void

  AUTHOR
  Dan Nguyen

  Diab Jerius

*/

void
str_tokenize_free(
  char** string_argv,	/* storage for the parsed tokens */
  int restore  		/* same as passed to str_tokenize */ )
{
  char **str_pptr = string_argv;

  if ( NULL == string_argv )
    return;
  
  if ( restore )
    while ( *str_pptr )
      free( *str_pptr++ );

  free( string_argv );
}
