/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of suplib
 *
 * suplib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * suplib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdio.h>
#include <stddef.h>
#include <errno.h>
#include <string.h>

#include "str/str.h"

/*
   PROCEDURE
   str_tokq - split a string at specified delimiters

   DESCRIPTION
   This routine works like @code{strtok}, except that it understands
   escaped characters (with @samp{\} as the escape character) and quoted
   strings.  It also can be told @emph{not} to skip over multiple
   consecutive delimiter characters (e.g., if the delimiter is @samp{,},
   the string ",,a" will return three tokens, the first two empty strings).
   Additionally @code{str_tokq} can be instructed to restore the parsed string
   to its initial state (normally it sticks @samp{\0} where the delimiter
   characters are).

   The invoking routine calls @code{str_tokq} multiple times, once per token.
   The first call should be made with @code{string} set to the address of the
   string to parse.  On subsequent calls it should be set to @code{NULL}.
   The delimiters may be changed on subsequent calls.

   Each invocation of @code{str_tokq} returns a pointer to the next
   token.  If no more tokens are available, it returns @code{NULL}.
   If skip mode is on (the parameter @code{skip} is non-zero), an
   empty string is considered to have no tokens.  If skip mode is off,
   an empty string is considered to have a single token.

   If the parameter @code{restore} is non-zero, @code{str_tokq} will restore
   the delimiter character which was replaced by a @samp{\0} on the
   @emph{previous} call.  Thus, to fully restore the string to its original
   state, @code{str_tokq} @emph{must} be invoked repeatedly until it
   returns a @code{NULL}.

   Escaped characters are recognized anywhere in a string.  Quotes
   need not be the first character in a token (e.g., @samp{foo" "bar})
   will be treated as one token, and can be either single (both
   forward and backwards) or double.  To embed quotes in like-quoted
   strings, escape them (e.g., @samp{foo"\""bar}).

   Unlike @code{strtok}, @code{str_tokq} is set up to allow concurrent use on
   multiple strings.  To accomplish this, all internal state information
   is kept in two user supplied locations.  The first (parameter @code{ptr}),
   is used by @code{strtok} to keep track of where it is in the string.  The
   second, @code{dchar}, contains the delimiter character that ended the
   token.  Please note that the invoking function must pass the
   @emph{addresses} of these locations to @code{str_tokq}.  Here's some
   sample code:
   @example
   char *ptr, dchar, *tok;
   tok = str_tokq(string, ",", 0, 1, &ptr, &dchar);
   while (tok)
   @{
      ... process tokens ...
      tok = str_tokq(NULL, ",", 0, 1, &ptr, &dchar);
   @}
   @end example

   ERRORS
   On error, @code{str_tokq} returns @code{NULL} and sets @code{errno}
   accordingly.   The following errors are recognized:
   
   @table @code
   @item EINVAL
     There were unbalanced quotes or the last character in the
     string was an escape character (i.e., no character to escape).

   @end table

*/
char *
str_tokq(
	char *string,	/* the string to parse */
	const char *delim,	/* the set of characters that delimit tokens */
	int skip,	/* true if consecutive delimiters are treated as
			   a single delimiter */
	int restore,	/* true if @code{str_tokq} should restore the
			   delimiter character from the previous invocation */
	char **ptr,	/* state information needed by @code{str_tokq} */
	char *dchar	/* state information needed by @code{str_tokq} */
)
{
  static char *start;
  char *estr;

  const char *dstart = delim;		/* temp to process delimiters */
  const char *dend = delim + strlen(delim);	/* temp to process delimiters */

  if (string != NULL)
  {
    start = string;
    *dchar = '\0';
  }
  else
  {
    if (*ptr == NULL)
      return NULL;
    
    start = *ptr;

    if (restore)
      *start = *dchar;

    start++;
  }
  
  /* skip over any delimiters at the front of the string */
  if (skip)
    start += strspn(start, delim);

  /* find the next delimiter */
  for (estr = start ; *estr ; estr++)
  {
    if (*estr == '\\')
    {
      if (*++estr == '\0')
      {
	errno = EINVAL;
	return NULL;
      }
    }
    else if (*estr == '"' || *estr == '\'' || *estr == '`')
    {
      char quote = *estr++;

      /* find its partner */
      for(;  *estr && *estr != quote ; estr++)
	if (*estr == '\\' && *++estr == '\0')
	  break;

      /* only a single quote. bogus dude! */
      if (*estr != quote)
      {
	errno = EINVAL;
	return NULL;
      }
    }
    else
    {
      /* see if the character is a delimiter */
      for (dstart = delim ; dstart < dend && *estr != *dstart ; dstart++ )
	;
      if (dstart != dend)
      {
	/* found delimiter */
	*dchar = *estr;
	*estr = '\0';
	*ptr = estr;
	return start;
      }
    }
  }

  /* this is reached upon hitting the end of the string */

  /*
     if skip is not turned on, even empty strings are treated
     as tokens.  however the default (strtok) behavior is to return a NULL.
     so, a little special processing for this case.  since
     estr is set to NULL, we'll only hit this code once. 
  */

  *ptr = NULL;
  return *start == '\0' && skip ? NULL : start;
}

/*
   PROCEDURE
   str_tokqcnt - count the number of tokens in a string

   DESCRIPTION
   This routine counts the number of tokens in a string as would be
   returned by @code{str_tokq}.  It @emph{does} change the input string,
   but restores it to its initial state upon completion.  A constant
   string (e.g., a statically declared string) should not be passed.

   RETURNS
   Upon success it returns the number of tokens in the string.
   On error it returns @samp{-1}.

   AUTHOR
   Diab Jerius

*/
int
str_tokqcnt(
  char *string,			/* the string to parse */
  const char *delim,		/* the set of characters that delimit tokens */
  int skip			/* true if consecutive delimiters are treated
				   as a single delimiter */
)
{
  int cnt = 0;
  char *ptr, dchar;
  char *tok;

  /* save errno */
  int errno_s = errno;
  errno = 0;

  tok = str_tokq(string, delim, skip, 1, &ptr, &dchar);
  while(tok)
  {
    cnt++;
    tok = str_tokq(NULL, delim, skip, 1, &ptr, &dchar);
  }

  if (errno == EINVAL)
    cnt = -1;

  errno = errno_s;

  return cnt;
}


/*
   PROCEDURE
   str_tokq_restore - restore string passed to str_tokq

   DESCRIPTION
   @code{str_tokq} changes the scanned string as it processes it.  If
   the @code{restore} flag passed to @code{str_tokq} is set then the
   string will be restored if the @code{str_tokq} is called to
   completion. However, sometimes it's necessary to interrupt the
   parsing, and repeatedly calling the @code{str_tokq} to restore the
   string is ridiculous.

   This function restores the string.

   RETURNS
   It returns a pointer to the next character that @code{str_tokq}
   would have returned, except in the case where the end of string was
   reached, in which case it will return a pointer to the end of
   string character.  This allows one to pass the pointer to another
   independent invocation of @code{str_tokq}, if,
   for example, continued parsing is to be done in another context.

   AUTHOR
   Diab Jerius

*/

char *
str_tokq_restore(
	char *ptr,	/* state information returned by @code{str_tokq} */
	char dchar	/* state information returned by @code{str_tokq} */
)
{
  if (ptr == NULL)
    return NULL;
    
  *ptr = dchar;

  return dchar ? ++ptr : ptr;
}
