#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>

#include <errno.h>

#include "str/str.h"
#include "debug/debug.h"

#include <testlib/testlib.h>

#define LABEL(extra) (sprintf( label, "%-34s: %s", prefix, extra ), label)


static
void test_tokq(
  char *string,
  const char *delim,
  int skip,
  int restore,
  int error,
  ...
  )
{
  char save[1024];
  char label[1024];
  char prefix[1024];

  char *buf;
  char *result;
  char *expected;
  char *ptr;
  char dchar;
  va_list ap;

  va_start( ap, error );

  sprintf( prefix, "%-13s: skip=%d restore=%d", string, skip, restore );
  
  /* make string r/w */
  strcpy( save, string );
  
  for( expected = (char*) va_arg(ap, char*),
	 buf = save ;
       expected ;
       expected = (char*) va_arg(ap, char*),
	 buf = NULL
       )
  {
    result = str_tokq( buf, delim, skip, restore, &ptr, &dchar );
    TESTSTRING( LABEL( expected ), result, expected );
    if ( strcmp( result, expected ) )
      fprintf( stderr, ">%s< <- >%s<\n", result, expected );
  }

  result = str_tokq( buf, delim, skip, restore, &ptr, &dchar );
  TEST( LABEL("NULL"), result, NULL );

  TEST( LABEL("error"), errno, error );

  if ( restore )
  {
    TESTSTRING( LABEL("restore"), save, string );
  }

  va_end(ap);

}


int main( void )
{
  char   buf[1024];
  int skip, restore;

  /*******************************************************************/
  START( "str_tokq" );

  /*#################################################################*/
  /* ideal input: no quotes, no skips */

  strcpy( buf, "a,b,c" );

  /******************************/
  skip = 0; restore = 1;

  test_tokq( buf, ",", skip, restore, 0, "a", "b", "c", NULL );

  /******************************/
  skip = 0; restore = 0;

  test_tokq( buf, ",", skip, restore, 0, "a", "b", "c", NULL );

  /******************************/
  skip = 1; restore = 1;

  test_tokq( buf, ",", skip, restore, 0, "a", "b", "c", NULL );

  /******************************/
  skip = 1; restore = 0;

  test_tokq( buf, ",", skip, restore, 0, "a", "b", "c", NULL );


  /*#################################################################*/
  /* ideal input: no quotes, skips */

  /******************************/
  skip = 0; restore = 1;
  strcpy( buf, ",a,b,,c,," );

  test_tokq( buf, ",", skip, restore, 0, "", "a", "b", "", "c", "", "", NULL );

  /******************************/
  skip = 0; restore = 0;

  test_tokq( buf, ",", skip, restore, 0, "", "a", "b", "", "c", "", "", NULL );

  /******************************/
  skip = 1; restore = 1;

  test_tokq( buf, ",", skip, restore, 0, "a", "b", "c", NULL );

  /******************************/
  skip = 1; restore = 0;

  test_tokq( buf, ",", skip, restore, 0, "a", "b", "c", NULL );


  /*#################################################################*/
  /* ideal input: quotes, no skips */

  strcpy( buf, "'a','b','c'" );

  /******************************/
  skip = 0; restore = 1;

  test_tokq( buf, ",", skip, restore, 0, "'a'", "'b'", "'c'", NULL );

  /******************************/
  skip = 0; restore = 0;

  test_tokq( buf, ",", skip, restore, 0, "'a'", "'b'", "'c'", NULL );

  /******************************/
  skip = 1; restore = 1;

  test_tokq( buf, ",", skip, restore, 0, "'a'", "'b'", "'c'", NULL );

  /******************************/
  skip = 1; restore = 0;

  test_tokq( buf, ",", skip, restore, 0, "'a'", "'b'", "'c'", NULL );

  /*#################################################################*/
  /* ideal input: quotes, skips */

  strcpy( buf, ",'a','b',,'c',," );

  /******************************/
  skip = 0; restore = 1;

  test_tokq( buf, ",", skip, restore, 0,
	     "", "'a'", "'b'", "", "'c'", "", "", NULL );

  /******************************/
  skip = 0; restore = 0;

  test_tokq( buf, ",", skip, restore, 0,
	     "", "'a'", "'b'", "", "'c'", "", "", NULL );

  /******************************/
  skip = 1; restore = 1;

  test_tokq( buf, ",", skip, restore, 0,
	     "'a'", "'b'", "'c'", NULL );

  /******************************/
  skip = 1; restore = 0;

  test_tokq( buf, ",", skip, restore, 0,
	     "'a'", "'b'", "'c'", NULL );


  /*#################################################################*/
  /* ideal input: escapes */
  strcpy( buf, "\\,'a','b',\\\\,c\\',," );

  /******************************/
  skip = 0; restore = 1;

  test_tokq( buf, ",", skip, restore, 0,
	     "\\,'a'", "'b'", "\\\\", "c\\'", "", "", NULL );


  /******************************/
  skip = 1; restore = 1;

  test_tokq( buf, ",", skip, restore, 0,
	     "\\,'a'", "'b'", "\\\\", "c\\'", NULL );


  
  /*#################################################################*/
  /* broken input: unbalanced quotes */
  strcpy( buf, "'a,'b',,,," );

  /******************************/
  skip = 0; restore = 1;

  test_tokq( buf, ",", skip, restore, EINVAL, NULL );

  /******************************/
  skip = 1; restore = 1;

  test_tokq( buf, ",", skip, restore, EINVAL, NULL );

  /*#################################################################*/
  /* broken input: more unbalanced quotes */
  strcpy( buf, "a,,'b,,," );

  /******************************/
  skip = 0; restore = 1;

  test_tokq( buf, ",", skip, restore, EINVAL, "a", "", NULL );

  /******************************/
  skip = 1; restore = 1;

  test_tokq( buf, ",", skip, restore, EINVAL, "a", NULL );


  /*#################################################################*/
  /* broken input: trailing escape */
  strcpy( buf, ",a,b,\\" );

  /******************************/
  skip = 0; restore = 1;

  test_tokq( buf, ",", skip, restore, EINVAL, "", "a", "b", NULL );

  /******************************/
  skip = 1; restore = 1;

  test_tokq( buf, ",", skip, restore, EINVAL, "a", "b", NULL );


  /*******************************************************************/
  SUMMARY();
  return FAILED( ) ? EXIT_FAILURE : EXIT_SUCCESS;


}
