/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of suplib
 *
 * suplib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * suplib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "config.h"

/* It is possible with strict enough GCC options to break /usr/include/sys/resource.h on RHEL 5.x & CentOS 5.x.
   Essentially __USE_GNU is turned off, and that results in a hanging comma in a structure definition.
   Tsk. Tsk.
 */
#ifdef HAVE_BROKEN_SYS_RESOURCE_H
/* Try to ascertain if this is the failure we expect */
#  ifdef __USE_GNU
#    error "/usr/include/sys/resource.h is broken and I don't know why!"
#  endif

#define __USE_GNU 1

#endif

#include <sys/resource.h>

/* restore things to what they were */
#ifdef HAVE_BROKEN_SYS_RESOURCE_H
#undef __USE_GNU
#endif


#include "times/times.h"

static Time Start;

static void
diff_timeval(struct timeval *diff, struct timeval *t1, struct timeval *t2)
{
  if (t1->tv_sec < t2->tv_sec ||
      (t1->tv_sec == t2->tv_sec && t1->tv_sec < t2->tv_sec))
  {
    struct timeval *temp = t1;
    t1 = t2;
    t2 = temp;
  }

  diff->tv_sec = t1->tv_sec - t2->tv_sec;
  if (t1->tv_usec < t2->tv_usec)
  {
    diff->tv_usec = t1->tv_usec + (1000000 - t2->tv_usec);
    diff->tv_sec--;
  }
  else
  {
    diff->tv_usec = t1->tv_usec - t2->tv_usec;
  }
}

static void
add_timeval(struct timeval *sum, struct timeval *t1, struct timeval *t2)
{
  sum->tv_sec = t1->tv_sec + t2->tv_sec;
  sum->tv_usec = t1->tv_usec + t2->tv_usec;
  while (sum->tv_usec >= 1000000)
  {
    sum->tv_usec -= 1000000;
    sum->tv_sec++;
  }
}

/*

  PROCEDURE
  clearTime - initialize a @code{Time} structure to zero.

  DESCRIPTION
  @code{clearTime} clears the passed @code{Time} structure.
  
  RETURNS
  It returns a pointer to the passed @code{Time} structure.

  AUTHOR
  Diab Jerius

*/
Time *
clearTime(
  Time *t		/* the Time to clear */
)
{
  if ( t )
  {
    t->user.tv_sec   = t->user.tv_usec   = 0;
    t->system.tv_sec = t->system.tv_usec = 0;
    t->clock.tv_sec  = t->clock.tv_usec  = 0;
  }

  return t;
}

/*

  PROCEDURE
  startTime - initialize a @code{Time} structure to the current time.

  DESCRIPTION
  @code{startTime} initializes an internal @code{Time} structure with
  the current user, system, and clock times for the process.  If the
  @code{start} parameter is not @code{NULL}, the same data are copied
  to the user specified @code{Time} structure.  The internal
  @code{Time} structure will be overwritten by each call to
  @code{startTime}.

  RETURNS
  It returns a pointer to the internal @code{Time} structure.

  AUTHOR
  Diab Jerius

*/
Time *
startTime(
  Time *start		/* the structure to fill. may be @code{NULL} */
)
{
  if (currentTime(&Start) == NULL) return NULL;
  if (start) *start = Start;

  return &Start;
}

/*

  PROCEDURE
  elapsedTime - return the time elapsed 

  DESCRIPTION
  @code{elapsedTime} determines the difference in time between the
  current user, system and clock times and either a passed @code{Time}
  structure, or the internal @code{Time} structure initialized by
  @code{startTime} (if the passed pointer @code{start} is @code{NULL}).

  RETURNS
  It returns a pointer to an internal @code{Time} structure containing
  the difference in times. This structure is private to @code{elapsedTime}
  and will be overwritten by subsequent calls to it.
  It returns NULL upon error.

  AUTHOR
  Diab Jerius

*/
Time *
elapsedTime(
  Time *start		/* the start time from which to determine
			   the elapsed time */
  )
{
  static Time Elapsed;
  Time Current;

  if (currentTime(&Current) == NULL) return NULL;
  if (start) diffTime(&Elapsed, &Current, start);
  else       diffTime(&Elapsed, &Current, &Start);

  return &Elapsed;
}

/*

  PROCEDURE
  currentTime - return the current process times

  DESCRIPTION
  @code{currentTime} fills a specified @code{Time} structure with the
  current user, system and clock times for the process.

  RETURNS
  It returns the passed pointer.

  AUTHOR
  Diab Jerius
*/
Time *
currentTime(
  Time *time		/* the destination @code{Time} structure */
  )
{
  struct rusage usage;
  
  if (time == NULL) return NULL;

  if ( getrusage( RUSAGE_SELF, &usage ) < 0 )
    return NULL;
  
  time->user = usage.ru_utime;
  time->system = usage.ru_stime;

  if ( gettimeofday( &time->clock, NULL ) < 0 )
    return NULL;

  return time;
}

/*
  PROCEDURE
  incTime - add some Time to some other Time

  DESCRIPTION
  @code{incTime} increments a @code{Time} structure by the
  amount in a another @code{Time} structure

  RETURNS
  It returns a pointer to the destination @code{Time} structure,
  @code{NULL} if either the destination or the increment was
  a @code{NULL} pointer.

  AUTHOR
  Diab Jerius

*/

Time *
incTime(
  Time *dest,		/* The Time which will be incremented */
  Time *inc		/* the Time increment */
  )
{
  static Time sum;

  if (dest == NULL || inc == NULL) return NULL;

  sum = *dest;

  add_timeval(&sum.user, &dest->user, &inc->user);
  add_timeval(&sum.system, &dest->system, &inc->system);
  add_timeval(&sum.clock, &dest->clock, &inc->clock);

  *dest = sum;

  return dest;
}

/*

  PROCEDURE
  diffTime - return the difference in times between two @code{Time} structures.

  DESCRIPTION
  @code{diffTime} fills an internal @code{Time} structure with the
  difference in times between two passed @code{Time} structures (
  @code{end} - @code{start}).  The internal structure will be
  overwritten by the next call to @code{diffTime}.  If the @code{diff}
  argument is not @code{NULL}, the difference data are copied there as
  well.

  RETURNS
  It returns a pointer to the internal @code{Time} structure upon success.
  If either of @code{start} or @code{end} are @code{NULL}, it returns
  @code{NULL}.

  AUTHOR
  Diab Jerius

*/
Time *
diffTime(
  Time *diff,		/* where to copy the time differences */
  Time *end,		/* the ending time */
  Time *start		/* the starting time */
  )
{
  static Time Diff;

  if (start == NULL || end == NULL) return NULL;

  diff_timeval(&Diff.user,   &end->user,   &start->user);
  diff_timeval(&Diff.system, &end->system, &start->system);
  diff_timeval(&Diff.clock,  &end->clock,  &start->clock);

  if (diff) *diff = Diff;
  return &Diff;
}

/*

  PROCEDURE
  stringTime - return a string representation the passed @code{Time} structure.

  DESCRIPTION
  @code{stringTime} generates a printable string representing the
  passed @code{Time} structure.  The string is rendered into internal
  storage which will be overwritten by then next call to @code{stringTime}.

  RETURNS
  It returns a pointer to a printable string.

  AUTHOR
  Diab Jerius

*/
char *
stringTime(Time *time)
{
  static char timebuf[512];
  double clk_tck = (double) sysconf(_SC_CLK_TCK);
  
  sprintf(timebuf, "u: %ld.%06ld s: %ld.%06ld c: %ld.%06ld",
	  time->user.tv_sec,   time->user.tv_usec,
	  time->system.tv_sec, time->system.tv_usec,
	  time->clock.tv_sec,  time->clock.tv_usec
	  );

  return timebuf;
}


